# Class to calculate Cross-in-Tray function
# Function taken from https://www.sfu.ca/~ssurjano/ackley.html
# May 3rd 2017
#
# Dr. Walter F. de Azevedo Jr.
# azevedolab.net
# May 3rd 2017
#

class Function3D(object):
    """Class to calculate a 3D function"""
    
    # Define constructor method
    def __init__(self):
        
        # Define attributes
        self.xmin = -10
        self.xmax = 10
        self.npoints = 300
        self.title = "Cross-in Tray Function"
        self.output = "cross_in_tray.png"
        
    # Define __str__ method
    def __str__(self):
        info = "\nross-in-Tray Function"
        info += "Global minima: f(1.3481,1.3491) = f(-1.3481,1.3491) = f(-1.3481,-1.3491) = f(1.3481,1.3491) = -2.06261"
        info += "Input domain: -10 <= x,y <= 10"
        return info
        
    # Define calc method
    def calc(self):
        """Method to calculate 3D function"""
        
        # Import library
        import numpy as np
        
        # Generate x and y axes
        x = np.outer(np.linspace(self.xmin, self.xmax, self.npoints), np.ones(self.npoints))
        y = x.copy().T
                
        # Generate f(x,y)
        f = -0.0001*( np.abs(np.sin(x)*np.sin(y))*np.exp(   np.abs(100 - (1/np.pi) *np.sqrt(x**2 + y**2))))**(0.1)
        
        # Return axes        
        return x,y,f,self.xmin,self.xmax,self.title,self.output