# Class to calculate Drop Wave Function
# Function taken from https://www.sfu.ca/~ssurjano/drop.html
# May 3rd 2017
#
# Dr. Walter F. de Azevedo Jr.
# azevedolab.net
# May 3rd 2017
#

class Function3D(object):
    """Class to calculate a 3D function"""
    
    # Define constructor method
    def __init__(self):
        
        # Define attributes
        self.xmin = -5.12
        self.xmax = 5.12
        self.npoints = 300
        self.title = "Drop Wave Function"
        self.output = "dropwave.png"
        
    # Define __str__ method
    def __str__(self):
        info = "\nDrop Wave Function"
        info += "Global minimum: f(0,0) = -1"
        info += "Input domain: -5.12 <= x, y <= 5.12"
        return info
        
    # Define calc method
    def calc(self):
        """Method to calculate 3D function"""
        
        # Import library
        import numpy as np
        
        # Generate x and y axes
        x = np.outer(np.linspace(self.xmin, self.xmax, self.npoints), np.ones(self.npoints))
        y = x.copy().T
                
        # Generate f(x,y)
        f = (-1 - np.cos(12*np.sqrt(x**2+y**2)) )/(0.5*(x**2+y**2)+2)
        
        # Return axes        
        return x,y,f,self.xmin,self.xmax,self.title,self.output