# Class to crelate Matplot3D objects
# It uses Matplotlib to generate Matplot3D objects
#
# Dr. Walter F. de Azevedo Jr.
# azevedolab.net
# May 3rd 2017
#
class Matplot3D(object):
    """Class to create 3D plot objects"""
    
    # Define constructor method
    def __init__(self,cs,rs,lwidth,color,bar):
        
        # Define attributes
        self.cs = cs            # cstride:	Array column stride (step size) (integer)
        self.rs = rs            # cstride:	Array column stride (step size) (integer)
        self.lwidth = lwidth    # linewidth (float)
        self.color = color      # Surface color (string)
        self.bar = bar          # Add a color bar if True (boolean)
    
    # Define __str__ method
    def __str__(self):
        info = "\nClass to generate a 3D plot using mpl_toolkits.mplot3d library"
        
        return info
            
    def plot_it(self,my_func):
        """Method to carry out plotting"""
            
        # Import libraries
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        import numpy as np
        
        # This method accepts the object into its parameter my_func. 
        # So, when plot_it() executes, my_func refers to a Function3D object
        # Invoking calc method from Function3D object
        x,y,f,self.xmin,self.xmax,self.title,self.output = my_func.calc()
        
        print("Generating Plot for "+self.title+"...")
        
        # Generate 3D plot
        fig = plt.figure()
        ax = plt.axes(projection='3d')
        
        # Check color
        if self.color =="cmap":
            surf = ax.plot_surface(x, y, f, cmap=plt.cm.jet, rstride=self.rs, cstride=self.cs, linewidth=self.lwidth)
        else:
            surf = ax.plot_surface(x, y, f, color = self.color, rstride=self.rs, cstride=self.cs, linewidth=self.lwidth)
        
        # Check bar to add a color bar
        if self.bar:
            # Add a color bar which maps values to colors.
            fig.colorbar(surf, shrink=0.5, aspect=5)
        
        # Include title
        plt.title(self.title)
        
        # Show plot
        plt.show()

        # Save figure
        plt.savefig(self.output)