# Class to calculate Rastrigin function
# Mathematical formula taken from https://en.wikipedia.org/wiki/Test_functions_for_optimization
# Accessed on May 3rd 2017
#
# Dr. Walter F. de Azevedo Jr.
# azevedolab.net
# May 3rd 2017
#

class Function3D(object):
    """Class to calculate a 3D function"""
    
    # Define constructor method
    def __init__(self):
        
        # Define attributes
        self.xmin = -5.12
        self.xmax = 5.12
        self.npoints = 300
        self.title = "Rastrigin Function"
        self.output = "rastrigin.png"
        
    # Define __str__ method
    def __str__(self):
        info = "\nRastrigin function"
        info += "Global minimum: f(0,0) = 0"
        info += "Search domain: -5.12 <= x,y <= 5.12"
        return info
        
    # Define calc method
    def calc(self):
        """Method to calculate 3D function"""
        
        # Import library
        import numpy as np
        
        # Generate x and y axes
        x = np.outer(np.linspace(self.xmin, self.xmax, self.npoints), np.ones(self.npoints))
        y = x.copy().T
                
        # Generate f(x,y)
        A = 10
        z = A + (x**2 - A*np.cos(2*np.pi*x)) + (y**2 - A*np.cos(2*np.pi*y))
        
        # Return axes        
        return x,y,z,self.xmin,self.xmax,self.title,self.output